/**
 * @addtogroup CBindings libmtime C language bindings
 * @{
 *
 * @file mtime_eventBasic.h
 *
 * @brief  Definition of the basic event type and its methods.
 *         Definition of Even-Groups which contains a list of events.
 *
 * @author  F. Prill, DWD.
 * @author  (Luis Kornblueh, Rahul Sinha, MPI-M).
 *
 * @date March 2020
 */


#ifndef _MTIME_EVENTBASIC_H
#define _MTIME_EVENTBASIC_H

#include <stdint.h>
#include <stdbool.h>
#include "mtime_timedelta.h"
#include "mtime_julianDay.h"


///provides a Maximum string length for Event names.
#define MAX_BASICEVENTNAME_STR_LEN 132

struct _datetime;
struct t_event_basic_cache;

/**
 * @struct event_basic
 *
 * @brief struct t_event_basic defines events. Events are set to be triggered at pre-specified intervals.
 *
 */
struct t_event_basic
{
  ///> Event's name.
  char*                         eventName;

  ///> is this event active?
  bool                          triggerCurrentEvent;

  ///> Anchor datetime. Can be NULL; Initialized to eventFirstDateTime.
  struct _datetime*             eventReferenceDateTime;
  ///> Start datetime of the event. Can be NULL; Initialized to 0-01-01T00:00:00.000.
  struct _datetime*             eventFirstDateTime;
  ///> Last datetime of the event. Can be NULL; stays a NULL pointer: Logically NULL.
  struct _datetime*             eventLastDateTime;
  ///> TimeDelta between successive triggers. MUST be specified.
  struct _timedelta*            eventInterval;

  ///> "Fuzziness interval" for event trigger comparisons. May be NULL.
  struct _timedelta*            plus_slack;
  ///> "Fuzziness interval" for event trigger comparisons. May be NULL.
  struct _timedelta*            minus_slack;

  ///> a number of data items which are derived from the data above.
  struct t_event_basic_cache     *cache;
};


/**
 * @struct t_event_basic_cache
 *
 * @brief A number of data items which are derived from the data above.
 *
 */
struct t_event_basic_cache
{
  struct _julianday                jd_eventReferenceDateTime;
  struct _juliandelta              jd_eventInterval;
  struct _divisionquotienttimespan quot_eventFirstDateTime;
  struct _divisionquotienttimespan quot_eventLastDateTime;

  /* quot_eventFirstDateTime rounded up. */
  struct _divisionquotienttimespan quot_eventFirstTriggerDateTime; 

  /* quot_eventLastDateTime rounded downwards. */
  struct _divisionquotienttimespan quot_eventLastTriggerDateTime;

  int64_t plus_slack_ms;
  int64_t minus_slack_ms;
};

int
basicEvent_cache_init(struct t_event_basic* e);

struct t_event_basic*
newBasicEventWithDataType(const char* eventName,
                          struct _datetime* eventReferenceDate,
                          struct _datetime* eventFirstDate,
                          struct _datetime* eventLastDate,
                          struct _timedelta* eventInterval,
                          struct _timedelta* plus_slack,
                          struct _timedelta* minus_slack);

void
deallocateBasicEvent(struct t_event_basic* e);

bool
isCurrentBasicEventActive(struct t_event_basic* e,
                          struct _datetime* current_dt,
                          struct _datetime* trigger_date);

bool
isBasicEventActiveInRange(struct t_event_basic* e,
                          struct _datetime* dt1,
                          struct _datetime* dt2);

char*
getBasicEventName(struct t_event_basic* e, char* ename);

struct _datetime*
getBasicEventReferenceDateTime(struct t_event_basic* e);

struct _datetime*
getBasicEventFirstDateTime(struct t_event_basic* e);

struct _datetime*
getBasicEventLastDateTime(struct t_event_basic* e);

struct _timedelta*
getBasicEventInterval(struct t_event_basic* e);

/**
 * @}
 */

#endif
